'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var invariant = _interopDefault(require('invariant'));
var $$observable = _interopDefault(require('symbol-observable'));
var most = require('most');

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          Promise.resolve(value).then(_next, _throw);
        }
      }

      function _next(value) {
        step("next", value);
      }

      function _throw(err) {
        step("throw", err);
      }

      _next();
    });
  };
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype.__proto__ = superClass && superClass.prototype;
  subClass.__proto__ = superClass;
}

function warning(message) {
  if (typeof console !== 'undefined' && typeof console.error === 'function') {
    console.error(message);
  }
  try {
    throw new Error(message);
  } catch (e) {
  }
}

function addToArray(a, b) {
  var i = a.indexOf(b);
  if (i === -1) {
    a.push(b);
  }
}
function removeFromArray(a, b) {
  var i = a.indexOf(b);
  if (i !== -1) {
    a.splice(i, 1);
  }
}
var _nextId = 0;
function nextId() {
  return _nextId++;
}
var unique = {
  equals: function equals(arg) {
    return false;
  },
  inspect: function inspect() {
    return 'Unique{_}';
  }
};
function defaultEquals(a, b) {
  return Object.is(a, b) || a && typeof a.equals === 'function' && a.equals(b);
}
function setEquals(derivable, eq) {
  derivable._equals = eq;
  return derivable;
}
function equals(ctx, a, b) {
  return (ctx._equals || defaultEquals)(a, b);
}

var parentsStack = [];
var child = null;
function startCapturingParents(_child, parents) {
  parentsStack.push({
    parents: parents,
    offset: 0,
    child: _child
  });
  child = _child;
}
function retrieveParentsFrame() {
  return parentsStack[parentsStack.length - 1];
}
function stopCapturingParents() {
  parentsStack.pop();
  child = parentsStack.length === 0 ? null : parentsStack[parentsStack.length - 1].child;
}
function maybeCaptureParent(p) {
  if (child === null) return;
  var frame = parentsStack[parentsStack.length - 1];
  if (frame.parents[frame.offset] === p) {
    frame.offset++;
    return;
  }
  var index = frame.parents.indexOf(p);
  if (index !== -1) {
    if (index <= frame.offset) return;
    var tmp = frame.parents[index];
    frame.parents[index] = frame.parents[frame.offset];
    frame.parents[frame.offset] = tmp;
    frame.offset++;
    return;
  }
  if (child !== undefined) {
    addToArray(p._activeChildren, child);
  }
  if (frame.offset === frame.parents.length) {
    frame.parents.push(p);
  } else {
    frame.parents.push(frame.parents[frame.offset]);
    frame.parents[frame.offset] = p;
  }
  frame.offset++;
}

var ATOM = 'ATOM';
var DERIVATION = 'DERIVATION';
var LENS = 'LENS';
var REACTOR = 'REACTOR';
function isDerivable(x) {
  return x !== undefined && x !== null && (x._type === DERIVATION || x._type === ATOM || x._type === LENS);
}

var UNKNOWN = 0;
var CHANGED = 1;
var UNCHANGED = 2;
var DISCONNECTED = 3;

function detach(parent, child) {
  removeFromArray(parent._activeChildren, child);
  if (parent._activeChildren.length === 0 && parent._parents != null) {
    var len = parent._parents.length;
    for (var i = 0; i < len; i++) {
      detach(parent._parents[i], parent);
    }
    parent._parents = null;
    parent._state = DISCONNECTED;
  }
}

var Reactor =
/*#__PURE__*/
function () {
  function Reactor(parent, react) {
    this._governor = null;
    this._parent = void 0;
    this.react = void 0;
    this._active = false;
    this._reacting = false;
    this._parent = parent;
    this.react = react;
  }
  var _proto = Reactor.prototype;
  _proto.start = function start() {
    this._active = true;
    addToArray(this._parent._activeChildren, this);
    this._parent.get();
  };
  _proto._force = function _force(nextValue) {
    this._reacting = true;
    try {
      this.react(nextValue);
    } finally {
      this._reacting = false;
    }
  };
  _proto.force = function force() {
    this._force(this._parent.get());
  };
  _proto.stop = function stop() {
    detach(this._parent, this);
    this._active = false;
  };
  _proto._maybeReact = function _maybeReact() {
    if (!this._reacting && this._active) {
      if (this._governor !== null) {
        this._governor._maybeReact();
      }
      if (this._active) {
        var nextValue = this._parent.get();
        if (this._parent._state === CHANGED) {
          this._force(nextValue);
        }
      }
    }
  };
  return Reactor;
}();
Object.defineProperty(Reactor.prototype, '_type', {
  value: REACTOR,
  configurable: true
});

function reactorFabric(derive, derivable, f, options) {
  !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'the first argument to .react must be a function') : invariant(false) : void 0;
  var opts = _extends({
    once: false,
    skipFirst: false
  }, options);
  var skipFirst = opts.skipFirst;
  var reactor = new Reactor(derivable, function (val) {
    if (skipFirst) {
      skipFirst = false;
    } else {
      f(val);
      if (opts.once) {
        reactor.stop();
        controller.stop();
      }
    }
  });
  var assertCondition = function assertCondition(condition, name) {
    if (isDerivable(condition)) {
      return condition;
    }
    if (typeof condition === 'function') {
      return condition;
    }
    if (typeof condition === 'undefined') {
      return condition;
    }
    throw Error("react " + name + " condition must be derivable or function, got: " + JSON.stringify(condition));
  };
  function getCondition(condition, def) {
    if (!condition) return def;
    if (typeof condition === 'function') return condition(derivable);
    return condition.get();
  }
  var $from = assertCondition(opts.from, 'from');
  var $until = assertCondition(opts.until, 'until');
  var $when = assertCondition(opts.when, 'when');
  var $conds = derive(function () {
    return {
      from: getCondition($from, true),
      until: getCondition($until, false),
      when: getCondition($when, true)
    };
  });
  var started = false;
  var controller = new Reactor($conds, function (conds) {
    if (conds.from) {
      started = true;
    }
    if (started) {
      if (conds.until) {
        reactor.stop();
        controller.stop();
      } else if (conds.when) {
        if (!reactor._active) {
          reactor.start();
          reactor.force();
        }
      } else if (reactor._active) {
        reactor.stop();
      }
    }
  });
  controller.start();
  controller.force();
  reactor._governor = controller;
}

function some(x) {
  return typeof x !== 'undefined' && x !== null;
}
function unpack(thing) {
  if (isDerivable(thing)) {
    return thing.get();
  } else {
    return thing;
  }
}
var Derivation =
/*#__PURE__*/
function () {
  function Derivation(deriver, meta) {
    if (meta === void 0) {
      meta = null;
    }
    this._parents = null;
    this._value = unique;
    this._equals = null;
    this._state = DISCONNECTED;
    this._deriver = void 0;
    this._meta = void 0;
    this._activeChildren = [];
    this._deriver = deriver;
    this._meta = meta;
  }
  var _proto = Derivation.prototype;
  _proto._clone = function _clone() {
    return setEquals(_derive(this._deriver), this._equals);
  };
  _proto._forceEval = function _forceEval() {
    var newVal = null;
    var newNumParents;
    try {
      if (this._parents === null) {
        this._parents = [];
      }
      startCapturingParents(this, this._parents);
      newVal = this._deriver();
      newNumParents = retrieveParentsFrame().offset;
    } finally {
      stopCapturingParents();
    }
    if (!equals(this, newVal, this._value)) {
      this._state = CHANGED;
    } else {
      this._state = UNCHANGED;
    }
    for (var i = newNumParents, len = this._parents.length; i < len; i++) {
      var oldParent = this._parents[i];
      detach(oldParent, this);
      this._parents[i] = null;
    }
    this._parents.length = newNumParents;
    this._value = newVal;
  };
  _proto._update = function _update() {
    if (this._parents === null) {
      this._forceEval();
    } else if (this._state === UNKNOWN) {
      var len = this._parents.length;
      for (var i = 0; i < len; i++) {
        var parent = this._parents[i];
        if (parent._state === UNKNOWN) {
          parent._update();
        }
        if (parent._state === CHANGED) {
          this._forceEval();
          break;
        }
      }
      if (this._state === UNKNOWN) {
        this._state = UNCHANGED;
      }
    }
  };
  _proto.get = function get() {
    maybeCaptureParent(this);
    if (this._activeChildren.length > 0) {
      this._update();
    } else {
      startCapturingParents(void 0, []);
      try {
        this._value = this._deriver();
      } finally {
        stopCapturingParents();
      }
    }
    return this._value;
  };
  _proto.derive = function derive(f) {
    var _this = this;
    !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'derive requires function') : invariant(false) : void 0;
    return _derive(function () {
      return f(_this.get());
    });
  };
  _proto.maybeDerive = function maybeDerive(f) {
    var _this2 = this;
    !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'maybeDerive requires function') : invariant(false) : void 0;
    return _derive(function () {
      var arg = _this2.get();
      return some(arg) ? f(arg) : null;
    });
  };
  _proto.orDefault = function orDefault(def) {
    !some(def) ? process.env.NODE_ENV !== "production" ? invariant(false, 'orDefault requires non-null value') : invariant(false) : void 0;
    return this.derive(function (value) {
      return some(value) ? value : def;
    });
  };
  _proto.react = function react(f, opts) {
    reactorFabric(_derive, this, f, opts);
  };
  _proto.maybeReact = function maybeReact(f, opts) {
    var maybeWhen = this.derive(Boolean);
    if (opts && 'when' in opts && opts.when !== true) {
      var when = opts.when;
      if (typeof when === 'function' || when === false) {
        when = _derive(when);
      }
      !isDerivable(when) ? process.env.NODE_ENV !== "production" ? invariant(false, 'when condition must be bool, function, or derivable') : invariant(false) : void 0;
      maybeWhen = maybeWhen.derive(function (d) {
        return d && when.get();
      });
    }
    reactorFabric(_derive, this, f, _extends({}, opts, {
      when: maybeWhen
    }));
  };
  _proto.is = function is(other) {
    var _this3 = this;
    return _derive(function () {
      return equals(_this3, _this3.get(), unpack(other));
    });
  };
  _proto.withEquality = function withEquality(equals$$1) {
    var selectedEquals = equals$$1;
    if (equals$$1) {
      !(typeof equals$$1 === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'equals must be function') : invariant(false) : void 0;
    } else {
      selectedEquals = null;
    }
    return setEquals(this._clone(), selectedEquals);
  };
  return Derivation;
}();
Object.defineProperty(Derivation.prototype, '_type', {
  value: DERIVATION,
  configurable: true
});
function _derive(f, meta) {
  !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'derive requires function') : invariant(false) : void 0;
  return new Derivation(f, meta);
}

function mark(node, reactors) {
  for (var i = 0, len = node._activeChildren.length; i < len; i++) {
    var child = node._activeChildren[i];
    switch (child._type) {
      case DERIVATION:
      case LENS:
        if (child._state !== UNKNOWN) {
          child._state = UNKNOWN;
          mark(child, reactors);
        }
        break;
      case REACTOR:
        reactors.push(child);
        break;
    }
  }
}
function processReactors(reactors) {
  for (var i = 0, len = reactors.length; i < len; i++) {
    var r = reactors[i];
    if (r._reacting) {
      throw Error('Synchronous cyclical reactions disallowed. ' + 'Use setImmediate.');
    }
    r._maybeReact();
  }
}
var TransactionAbortion = {};
function initiateAbortion() {
  throw TransactionAbortion;
}
var TransactionContext = function TransactionContext(parent) {
  this.id2originalValue = {};
  this.parent = void 0;
  this.modifiedAtoms = [];
  this.parent = parent;
};
function maybeTrack(atom) {
  if (currentCtx === null) return;
  if (atom._id in currentCtx.id2originalValue) return;
  currentCtx.modifiedAtoms.push(atom);
  currentCtx.id2originalValue[atom._id] = atom._value;
}
var currentCtx = null;
function inTransaction() {
  return currentCtx !== null;
}
function transact(f) {
  beginTransaction();
  try {
    f(initiateAbortion);
  } catch (e) {
    abortTransaction();
    if (e !== TransactionAbortion) {
      throw e;
    }
    return;
  }
  commitTransaction();
}
function atomically(f) {
  if (!inTransaction()) {
    transact(f);
  } else {
    f();
  }
}
function beginTransaction() {
  currentCtx = new TransactionContext(currentCtx);
}
function commitTransaction() {
  if (currentCtx === null) return;
  var ctx = currentCtx;
  currentCtx = ctx.parent;
  if (currentCtx !== null) return;
  var reactors = [];
  ctx.modifiedAtoms.forEach(function (a) {
    if (equals(a, a._value, ctx.id2originalValue[a._id])) {
      a._state = UNCHANGED;
    } else {
      a._state = CHANGED;
      mark(a, reactors);
    }
  });
  processReactors(reactors);
  ctx.modifiedAtoms.forEach(function (a) {
    a._state = UNCHANGED;
  });
}
function abortTransaction() {
  if (currentCtx === null) return;
  var ctx = currentCtx;
  currentCtx = ctx.parent;
  ctx.modifiedAtoms.forEach(function (atom) {
    atom._value = ctx.id2originalValue[atom._id];
    atom._state = UNCHANGED;
    mark(atom, []);
  });
}

function makeReactor(derivable, f, options) {
  return reactorFabric(_derive, derivable, f, options);
}

function update(instance, f, args) {
  switch (args.length) {
    case 0:
      return instance.set(f(instance.get()));
    case 1:
      return instance.set(f(instance.get(), args[0]));
    case 2:
      return instance.set(f(instance.get(), args[0], args[1]));
    case 3:
      return instance.set(f(instance.get(), args[0], args[1], args[2]));
    case 4:
      return instance.set(f(instance.get(), args[0], args[1], args[2], args[3]));
    default:
      throw Error('update function accepts only 4 arguments');
  }
}

function unpack$1(thing) {
  if (isDerivable(thing)) {
    return thing.get();
  } else {
    return thing;
  }
}

function some$1(x) {
  return typeof x !== 'undefined' && x !== null;
}
var Atom =
/*#__PURE__*/
function () {
  function Atom(value, meta) {
    if (meta === void 0) {
      meta = null;
    }
    this._id = nextId();
    this._state = UNCHANGED;
    this._equals = null;
    this._value = void 0;
    this._meta = void 0;
    this._activeChildren = [];
    this._value = value;
    this._meta = meta;
  }
  var _proto = Atom.prototype;
  _proto.derive = function derive$$1(f) {
    var _this = this;
    !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'derive requires function') : invariant(false) : void 0;
    return _derive(function () {
      return f(_this.get());
    });
  };
  _proto.maybeDerive = function maybeDerive(f) {
    var _this2 = this;
    !(typeof f === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'maybeDerive requires function') : invariant(false) : void 0;
    return _derive(function () {
      var arg = _this2.get();
      return some$1(arg) ? f(arg) : null;
    });
  };
  _proto.orDefault = function orDefault(def) {
    !some$1(def) ? process.env.NODE_ENV !== "production" ? invariant(false, 'orDefault requires non-null value') : invariant(false) : void 0;
    return this.derive(function (value) {
      return some$1(value) ? value : def;
    });
  };
  _proto.react = function react(f, opts) {
    makeReactor(this, f, opts);
  };
  _proto.maybeReact = function maybeReact(f, opts) {
    var maybeWhen = this.derive(Boolean);
    if (opts && 'when' in opts && opts.when !== true) {
      var when = opts.when;
      if (typeof when === 'function' || when === false) {
        when = _derive(when);
      }
      !isDerivable(when) ? process.env.NODE_ENV !== "production" ? invariant(false, 'when condition must be bool, function, or derivable') : invariant(false) : void 0;
      maybeWhen = maybeWhen.derive(function (d) {
        return d && when && when.get();
      });
    }
    makeReactor(this, f, _extends({}, opts, {
      when: maybeWhen
    }));
  };
  _proto.is = function is(other) {
    var _this3 = this;
    return _derive(function () {
      return equals(_this3, _this3.get(), unpack$1(other));
    });
  };
  _proto.withEquality = function withEquality(equals$$1) {
    var eq = equals$$1;
    if (eq) {
      !(typeof equals$$1 === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'equals must be function') : invariant(false) : void 0;
    } else {
      eq = null;
    }
    return setEquals(this._clone(), eq);
  };
  _proto._clone = function _clone() {
    return setEquals(atom(this._value), this._equals);
  };
  _proto.set = function set(value) {
    maybeTrack(this);
    var oldValue = this._value;
    this._value = value;
    if (!inTransaction()) {
      if (!equals(this, value, oldValue)) {
        try {
          this._state = CHANGED;
          var reactors = [];
          mark$1(this, reactors);
          processReactors(reactors);
        } finally {
          this._state = UNCHANGED;
        }
      }
    }
  };
  _proto.get = function get() {
    if (typeof global.__DERIVABLE_DEVTOOLS_HOOK__ === 'function') {
      global.__DERIVABLE_DEVTOOLS_HOOK__('captureAtom', this);
    }
    maybeCaptureParent(this);
    return this._value;
  };
  _proto.update = function update$$1(f) {
    for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }
    return update(this, f, args);
  };
  return Atom;
}();
Object.defineProperty(Atom.prototype, '_type', {
  value: ATOM,
  configurable: true
});
function atom(value, meta) {
  return new Atom(value, meta);
}
function mark$1(node, reactors) {
  for (var i = 0, len = node._activeChildren.length; i < len; i++) {
    var child = node._activeChildren[i];
    switch (child._type) {
      case DERIVATION:
      case LENS:
        if (child._state !== UNKNOWN) {
          child._state = UNKNOWN;
          mark$1(child, reactors);
        }
        break;
      case REACTOR:
        reactors.push(child);
        break;
    }
  }
}

var Lens =
/*#__PURE__*/
function (_Derivation) {
  function Lens(descriptor, meta) {
    var _this;
    _this = _Derivation.call(this, descriptor.get, meta) || this;
    _this._descriptor = void 0;
    _this._descriptor = descriptor;
    return _this;
  }
  var _proto = Lens.prototype;
  _proto._clone = function _clone() {
    return setEquals(new Lens(this._descriptor), this._equals);
  };
  _proto.set = function set(value) {
    var _this2 = this;
    atomically(function () {
      _this2._descriptor.set(value);
    });
  };
  _proto.update = function update$$1(f) {
    for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }
    return update(this, f, args);
  };
  _inheritsLoose(Lens, _Derivation);
  return Lens;
}(Derivation);
Object.defineProperty(Lens.prototype, '_type', {
  value: LENS,
  configurable: true
});

function createEvent(name) {
  return eventConstructor({
    name: name,
    domainName: ''
  });
}
function createEffect(name) {
  return effectConstructor({
    name: name,
    domainName: ''
  });
}
function eventConstructor(_ref) {
  var name = _ref.name,
      domainName = _ref.domainName;
  var fullName = makeName(name, domainName);
  var eventState = atom({
    payload: null
  });
  var instance = function instance(payload) {
    return instance.create(payload, fullName);
  };
  setProperty('create', create, instance);
  setProperty('eventState', eventState, instance);
  setProperty('toString', getType, instance);
  setProperty('getType', getType, instance);
  setProperty('kind', function () {
    return 'event';
  }, instance);
  setProperty($$observable, function () {
    return instance;
  }, instance);
  instance.watch = watch;
  instance.map = map;
  instance.subscribe = subscribe;
  instance.to = to;
  instance.epic = epic;
  function epic(fn) {
    var instance$ = most.from(instance).multicast();
    var epic$ = fn(instance$).multicast();
    var mapped = eventConstructor({
      name: name + "$ ~> *",
      domainName: domainName
    });
    epic$.observe(mapped);
    return mapped;
  }
  function to(target, handler) {
    !(target.kind() === 'store') ? process.env.NODE_ENV !== "production" ? invariant(false, 'right now event.to support only stores') : invariant(false) : void 0;
    !(typeof target.setState === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'right now event.to support only stores') : invariant(false) : void 0;
    watch(function (payload) {
      return target.setState(payload, handler);
    });
  }
  function create(payload, fullName) {
    instance.eventState.set({
      payload: payload
    });
    return payload;
  }
  function watch(watcher) {
    instance.eventState.react(function () {
      watcher(instance.eventState.get().payload);
    }, {
      skipFirst: true
    });
  }
  function subscribe(observer) {
    watch(function (payload) {
      return observer.next(payload);
    });
    function unsubscribe() {
      process.env.NODE_ENV !== "production" ? warning('TODO event.unsubscribe is not implemented') : void 0;
    }
    unsubscribe.unsubscribe = unsubscribe;
    return unsubscribe;
  }
  function map(fn) {
    var mapped = eventConstructor({
      name: name + " \u2192 *",
      domainName: domainName
    });
    watch(function (payload) {
      mapped(fn(payload));
    });
    return mapped;
  }
  function getType() {
    return fullName;
  }
  return instance;
}
function effectConstructor(_ref2) {
  var name = _ref2.name,
      domainName = _ref2.domainName;
  var instance = eventConstructor({
    name: name,
    domainName: domainName
  });
  var eventCreate = instance.create;
  var done = eventConstructor({
    name: name + ' done',
    domainName: domainName
  });
  var fail = eventConstructor({
    name: name + ' fail',
    domainName: domainName
  });
  instance.done = done;
  instance.fail = fail;
  instance.use = use;
  setProperty('create', create, instance);
  setProperty('kind', function () {
    return 'effect';
  }, instance);
  var thunk =
  /*#__PURE__*/
  function () {
    var _ref3 = _asyncToGenerator(
    /*#__PURE__*/
    regeneratorRuntime.mark(function _callee(value) {
      return regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              process.env.NODE_ENV !== "production" ? warning('no thunk used') : void 0;
            case 1:
            case "end":
              return _context.stop();
          }
        }
      }, _callee, this);
    }));
    return function thunk(_x) {
      return _ref3.apply(this, arguments);
    };
  }();
  function use(fn) {
    thunk = fn;
  }
  function create(payload) {
    var throwSymbol = {};
    var syncError = throwSymbol;
    var req;
    try {
      req = thunk(payload);
    } catch (err) {
      syncError = err;
    }
    var isSyncError = syncError !== throwSymbol;
    var isPromise = !isSyncError && hasPromise(req);
    if (isPromise) req.then(done, fail);else if (isSyncError) fail(syncError);else done(req);
    eventCreate(payload, instance.getType());
    return {
      done: function done() {
        return new Promise(function (rs) {
          if (isPromise) req.then(rs);else if (!isSyncError) rs(req);
        });
      },
      fail: function fail() {
        return new Promise(function (rs) {
          if (isPromise) req.catch(rs);else if (isSyncError) rs(syncError);
        });
      },
      promise: function promise() {
        if (isPromise) return req;
        if (isSyncError) return Promise.reject(syncError);
        return Promise.resolve(req);
      }
    };
  }
  return instance;
}
function hasPromise(req) {
  return typeof req === 'object' && req !== null && typeof req.then === 'function';
}
function makeName(name, domainName) {
  return [name, domainName].filter(function (str) {
    return str.length > 0;
  }).join('/');
}
function setProperty(prop, value, obj) {
  Object.defineProperty(obj, prop, {
    value: value,
    writable: true,
    configurable: true
  });
}

var INIT = '@@redux/INIT' + Math.random().toString(36).substring(7).split('').join('.');
var REPLACE = '@@redux/REPLACE' + Math.random().toString(36).substring(7).split('').join('.');

function compose() {
  for (var _len = arguments.length, funcs = new Array(_len), _key = 0; _key < _len; _key++) {
    funcs[_key] = arguments[_key];
  }
  if (funcs.length === 0) {
    return function (arg) {
      return arg;
    };
  }
  if (funcs.length === 1) {
    return funcs[0];
  }
  return funcs.reduce(function (a, b) {
    return function () {
      return a(b.apply(void 0, arguments));
    };
  });
}

function applyMiddleware() {
  for (var _len = arguments.length, middlewares = new Array(_len), _key = 0; _key < _len; _key++) {
    middlewares[_key] = arguments[_key];
  }
  return function (createStore) {
    return function () {
      var store = createStore.apply(void 0, arguments);
      var _dispatch = function dispatch() {
        throw new Error("Dispatching while constructing your middleware is not allowed. " + "Other middleware would not be applied to this dispatch.");
      };
      var chain = [];
      var middlewareAPI = {
        getState: store.getState,
        dispatch: function dispatch() {
          return _dispatch.apply(void 0, arguments);
        }
      };
      chain = middlewares.map(function (middleware) {
        return middleware(middlewareAPI);
      });
      _dispatch = compose.apply(void 0, chain)(store.dispatch);
      return _extends({}, store, {
        dispatch: _dispatch
      });
    };
  };
}

var _marked =
/*#__PURE__*/
regeneratorRuntime.mark(untilEnd);
function untilEnd(set) {
  var _iterator, _isArray, _i, _ref, e;
  return regeneratorRuntime.wrap(function untilEnd$(_context) {
    while (1) {
      switch (_context.prev = _context.next) {
        case 0:
          _iterator = set, _isArray = Array.isArray(_iterator), _i = 0, _iterator = _isArray ? _iterator : _iterator[Symbol.iterator]();
        case 1:
          if (!_isArray) {
            _context.next = 7;
            break;
          }
          if (!(_i >= _iterator.length)) {
            _context.next = 4;
            break;
          }
          return _context.abrupt("break", 17);
        case 4:
          _ref = _iterator[_i++];
          _context.next = 11;
          break;
        case 7:
          _i = _iterator.next();
          if (!_i.done) {
            _context.next = 10;
            break;
          }
          return _context.abrupt("break", 17);
        case 10:
          _ref = _i.value;
        case 11:
          e = _ref;
          set.delete(e);
          _context.next = 15;
          return e;
        case 15:
          _context.next = 1;
          break;
        case 17:
          if (set.size > 0) {
            _context.next = 0;
            break;
          }
        case 18:
        case "end":
          return _context.stop();
      }
    }
  }, _marked, this);
}
var id = 0;
function storeConstructor(props) {
  var _store;
  var currentId = (++id).toString(36);
  var currentListeners = [];
  var pending = new Set();
  var isDispatching = false;
  var currentReducer = props.currentReducer,
      currentState = props.currentState;
  var nextListeners = currentListeners;
  var nests = getNested(currentState, function (fn) {
    if (!isDispatching) return setState(fn(currentState));
    pending.add(fn);
  });
  var defaultState = currentState;
  var needToSaveFirst = defaultState === undefined;
  nests.add({
    get: function get() {
      return getState();
    },
    set: function set(state, action) {
      return currentReducer(state, action);
    }
  });
  function ensureCanMutateNextListeners() {
    if (nextListeners === currentListeners) {
      nextListeners = currentListeners.slice();
    }
  }
  function getState() {
    !!isDispatching ? process.env.NODE_ENV !== "production" ? invariant(false, 'You may not call store.getState() while the reducer is executing. ' + 'The reducer has already received the state as an argument. ' + 'Pass it down from the top reducer instead of reading it from the store.') : invariant(false) : void 0;
    return currentState;
  }
  function subscribe(listener) {
    !(typeof listener === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the listener to be a function.') : invariant(false) : void 0;
    !!isDispatching ? process.env.NODE_ENV !== "production" ? invariant(false, 'You may not call store.subscribe() while the reducer is executing. ' + 'If you would like to be notified after the store has been updated, subscribe from a ' + 'component and invoke store.getState() in the callback to access the latest state') : invariant(false) : void 0;
    var isSubscribed = true;
    ensureCanMutateNextListeners();
    nextListeners.push(listener);
    return function unsubscribe() {
      if (!isSubscribed) {
        return;
      }
      !!isDispatching ? process.env.NODE_ENV !== "production" ? invariant(false, 'You may not unsubscribe from a store listener while the reducer is executing') : invariant(false) : void 0;
      isSubscribed = false;
      ensureCanMutateNextListeners();
      var index = nextListeners.indexOf(listener);
      nextListeners.splice(index, 1);
    };
  }
  function dispatch(action) {
    !(typeof action.type !== 'undefined') ? process.env.NODE_ENV !== "production" ? invariant(false, 'Actions may not have an undefined "type" property.') : invariant(false) : void 0;
    !!isDispatching ? process.env.NODE_ENV !== "production" ? invariant(false, 'Reducers may not dispatch actions.') : invariant(false) : void 0;
    isDispatching = true;
    try {
      currentState = setNested(currentState, action, nests);
      for (var _iterator2 = untilEnd(pending), _isArray2 = Array.isArray(_iterator2), _i2 = 0, _iterator2 = _isArray2 ? _iterator2 : _iterator2[Symbol.iterator]();;) {
        var _ref2;
        if (_isArray2) {
          if (_i2 >= _iterator2.length) break;
          _ref2 = _iterator2[_i2++];
        } else {
          _i2 = _iterator2.next();
          if (_i2.done) break;
          _ref2 = _i2.value;
        }
        var fn = _ref2;
        currentState = fn(currentState);
      }
    } finally {
      if (needToSaveFirst === true && currentState !== undefined) {
        needToSaveFirst = false;
        defaultState = currentState;
      }
      isDispatching = false;
    }
    var listeners = currentListeners = nextListeners;
    for (var i = 0; i < listeners.length; i++) {
      var listener = listeners[i];
      listener(currentState, action);
    }
    return action;
  }
  function replaceReducer(nextReducer) {
    !(typeof nextReducer === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the nextReducer to be a function.') : invariant(false) : void 0;
    currentReducer = nextReducer;
    dispatch({
      type: REPLACE
    });
  }
  function observable() {
    var _ref3;
    var outerSubscribe = subscribe;
    return _ref3 = {
      subscribe: function subscribe(observer) {
        !(typeof observer === 'object' && observer !== null) ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the observer to be an object.') : invariant(false) : void 0;
        function observeState() {
          if (observer.next) {
            observer.next(getState());
          }
        }
        observeState();
        var unsubscribe = outerSubscribe(observeState);
        return {
          unsubscribe: unsubscribe
        };
      }
    }, _ref3[$$observable] = function () {
      return this;
    }, _ref3;
  }
  function reset(event) {
    on(event, function () {
      setState(defaultState);
    });
    return store;
  }
  function on(event, handler) {
    event.to(store, handler);
    return store;
  }
  function withProps(fn) {
    return function (props) {
      return fn(getState(), props);
    };
  }
  function map(fn) {
    var innerStore = createStore(fn(getState()));
    var unsub = watch(function (update) {
      var mapped = fn(update);
      innerStore.setState(mapped);
    });
    return innerStore;
  }
  function to(action, reduce) {
    var needReduce = action.kind() === 'store' && typeof reduce === 'function';
    return watch(function (data) {
      if (!needReduce) {
        action(data);
      } else {
        var lastState = action.getState();
        var reduced = reduce(lastState, data);
        if (lastState !== reduced) action.setState(reduced);
      }
    });
  }
  function watch(fn) {
    return subscribe(function () {
      fn(getState());
    });
  }
  function epic(fn) {
    return epicStore(store, fn);
  }
  function stateSetter(_, payload) {
    return payload;
  }
  function setState(value, reduce) {
    var currentReducer = typeof reduce === 'function' ? reduce : stateSetter;
    currentState = currentReducer(currentState, value);
    dispatch({
      type: "set state " + currentId,
      payload: value
    });
  }
  dispatch({
    type: INIT
  });
  var store = (_store = {
    id: currentId,
    withProps: withProps,
    setState: setState,
    dispatch: dispatch,
    map: map,
    on: on,
    to: to,
    watch: watch,
    epic: epic,
    subscribe: subscribe,
    getState: getState,
    replaceReducer: replaceReducer,
    reset: reset
  }, _store[$$observable] = observable, _store);
  Object.defineProperty(store, 'kind', {
    writable: true,
    configurable: true,
    value: function value() {
      return 'store';
    }
  });
  return store;
}
function epicStore(store, fn) {
  var store$ = most.from(store).multicast();
  var mapped$ = fn(store$).multicast();
  var innerStore = createStore();
  var subs = mapped$.subscribe({
    next: function next(value) {
      innerStore.setState(value);
    },
    error: function error(err) {
      if (process.env.NODE_ENV !== "production") console.error(err);
    },
    complete: function complete() {
      subs();
    }
  });
  return innerStore;
}
function hasKind(value) {
  return typeof value === 'object' && value !== null && typeof value.kind === 'function';
}
function createStore(preloadedStateRaw, reducer, enhancerRaw) {
  if (reducer === void 0) {
    reducer = function reducer(_) {
      return _;
    };
  }
  var enhancer = enhancerRaw;
  var preloadedState = preloadedStateRaw;
  if (typeof preloadedState === 'function' && typeof enhancer === 'undefined') {
    enhancer = preloadedState;
    preloadedState = undefined;
  }
  if (typeof enhancer !== 'undefined') {
    if (typeof enhancer !== 'function') {
      !Array.isArray(enhancer) ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the enhancer to be a function') : invariant(false) : void 0;
      enhancer = applyMiddleware.apply(void 0, enhancer);
    }
    return enhancer(createStore)(reducer, preloadedState);
  }
  !(typeof reducer === 'function') ? process.env.NODE_ENV !== "production" ? invariant(false, 'Expected the reducer to be a function') : invariant(false) : void 0;
  return storeConstructor({
    currentReducer: reducer,
    currentState: preloadedState
  });
}
function setNested(initialState, action, nests) {
  var currentState = initialState;
  for (var _iterator3 = nests, _isArray3 = Array.isArray(_iterator3), _i3 = 0, _iterator3 = _isArray3 ? _iterator3 : _iterator3[Symbol.iterator]();;) {
    var _ref4;
    if (_isArray3) {
      if (_i3 >= _iterator3.length) break;
      _ref4 = _iterator3[_i3++];
    } else {
      _i3 = _iterator3.next();
      if (_i3.done) break;
      _ref4 = _i3.value;
    }
    var _nest = _ref4;
    currentState = _nest.set(currentState, action);
  }
  return currentState;
}
function nest(value, key) {
  return {
    get: function get() {
      return value.getState();
    },
    set: function set(state, action) {
      value.dispatch(action);
      state[key] = value.getState();
      return state;
    }
  };
}
function getNested(initialState, setState) {
  var nests = new Set();
  if (typeof initialState !== 'object' || initialState === null) return nests;
  var _arr = Object.entries(_extends({}, initialState));
  var _loop = function _loop() {
    var _arr$_i = _arr[_i4],
        key = _arr$_i[0],
        value = _arr$_i[1];
    if (hasKind(value) && value.kind() === 'store') {
      var n = nest(value, key);
      nests.add(n);
      value.watch(function (e) {
        setState(function (state) {
          var _extends2;
          return _extends({}, state, (_extends2 = {}, _extends2[key] = e, _extends2));
        });
      });
      initialState[key] = value.getState();
    }
  };
  for (var _i4 = 0; _i4 < _arr.length; _i4++) {
    _loop();
  }
  return nests;
}

function createDomain(_ref) {
  var _domain = _ref.domain;
  return {
    event: function event(name) {
      return createEvent(name);
    },
    effect: function effect(name) {
      return createEffect(name);
    },
    domain: function domain(name) {
      return createDomain({
        domain: [name].concat(_domain)
      });
    },
    store: function store(state) {
      return createStore(state);
    }
  };
}
function createRootDomain(name) {
  if (name === undefined) return createDomain({
    domain: []
  });
  return createDomain({
    domain: [name]
  });
}

var defaultDomain = createRootDomain();
function createEvent$1(name) {
  return defaultDomain.event(name);
}
function createEffect$1(name) {
  return defaultDomain.effect(name);
}

exports.createDomain = createRootDomain;
exports.createStore = createStore;
exports.createEvent = createEvent$1;
exports.createEffect = createEffect$1;
//# sourceMappingURL=effector.cjs.js.map
